/* DUKE plugin for setup */
/* $Id: tar.c,v 1.12 2004/11/02 03:48:57 megastep Exp $ */


/* This is a total hack.
 *
 * This takes a file, verifies it's the install archive from the original
 *  MacSoft Duke3D retail disc, and then pull duke3d.grp out of it.
 *
 * We don't fully understand the format of the archive, but got enough of
 *  it to extract the .grp file. The rest we can safely package
 *  without inviting a lawsuit.  :)
 */

#include "config.h"
#include "plugins.h"
#include "file.h"
#include "md5.h"
#include "install_log.h"

#include <stdlib.h>
#include <string.h>
#ifdef HAVE_STRINGS_H
# include <strings.h>
#endif
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>


#define COMPSIZE 20805557
#define DECOMPSIZE 44356548

void cpt_wrfile1(unsigned char *in_char, unsigned long ibytes,
                 unsigned long obytes, unsigned long blocksize,
                 int type);

void (*cpt_update_hook)(int pos, int total) = NULL;


static UIUpdateFunc updatefunc = NULL;
static install_info *staticinfo = NULL;
static const char *staticfinal = NULL;
static const char *staticoption = NULL;
static size_t startbytes = 0;

static void duke_update(int pos, int total)
{
    if ( updatefunc ) {
        float pct = (((float) pos) / ((float) total)) * 100.0f;
	    updatefunc(staticinfo, staticfinal, (int) ((pct * 0.33f) + 33.0f), 100, staticoption);
    }

    staticinfo->installed_bytes = startbytes + (pos / 3);
}

static void decompress(unsigned char *src, int srclen, unsigned char *dst, int dstlen, int type)
{
	extern unsigned char *out_ptr;
	out_ptr = dst;
    cpt_update_hook = duke_update;
	cpt_wrfile1(src, srclen, dstlen, type, 0x1fff0);
}


/* Initialize the plugin */
static int DukeInitPlugin(void)
{
	return 1;
}

/* Free the plugin */
static int DukeFreePlugin(void)
{
	return 1;
}


static unsigned int read_int32(install_info *info, stream *io)
{
    unsigned char buf[4];
    if (file_read(info, buf, 4, io) != 4) return 0;
    return( (((unsigned int)buf[0])<<24) |
            (((unsigned int)buf[1])<<16) |
            (((unsigned int)buf[2])<<8)  |
            (((unsigned int)buf[3])<<0)   );
}

static int valid_archive(install_info *info, const char *path, stream *io)
{
    char buf[256];
    int pos = 0;
    int fsize = 0;

    /* Find start of duke3d.grp's record. */
    fsize = file_size(info, path);
    if (fsize == 24173847)        /* First US disc revision? */
        fseek(io->fp, 1100, SEEK_SET);
    else if (fsize == 24194283)   /* Second US disc revision? */
        fseek(io->fp, 1161, SEEK_SET);
    else if (fsize == 23910407)   /* French disc */
        fseek(io->fp, 1100, SEEK_SET);
    else
        return 0;  /* unknown disc. */

    if (file_read(info, buf, 1, io) != 1) return 0;
    if (buf[0] != 12) return 0;
    if (file_read(info, buf, 13, io) != 13) return 0;  /* intentionally read one extra. */
    if (memcmp(buf, "Duke3D Group", 12) != 0) return 0;
    pos = read_int32(info, io);  /* offset for compressed data */
    read_int32(info, io); /* "data" */
    read_int32(info, io); /* "Duke" */
    read_int32(info, io); /* creation date? */
    read_int32(info, io); /* mod date? */
    fseek(io->fp, 2, SEEK_CUR); /* skip 2 byte field. */
    if (read_int32(info, io) != 0x02C2300E) return 0; /* CRC */
    fseek(io->fp, 2, SEEK_CUR); /* skip 2 byte field. */
    read_int32(info, io); /* rsrc uncompressed size */
    if (read_int32(info, io) != DECOMPSIZE) return 0; /* data uncompressed size */
    read_int32(info, io); /* rsrc compressed size */
    if (read_int32(info, io) != COMPSIZE) return 0; /* data compressed size */

    fseek(io->fp, 0, SEEK_SET);
    file_skip(info, pos, io);  /* location of compressed duke3d.grp. */

    return 1;
}

/* Get the size of the file */
static size_t DukeSize(install_info *info, const char *path)
{
	log_debug("DUKE: Size(%s)", path);
	return DECOMPSIZE;  // actual size of atomic edition duke3d.grp ...
}

/* Extract the file */
static size_t DukeCopy(install_info *info, const char *path, const char *dest, const char *current_option,
		      xmlNodePtr node,
		      UIUpdateFunc update)
{
    static char *inbuf = NULL;
    static char *outbuf = NULL;
    stream *input, *output;
    char final[BUFSIZ];
    int br = 0;
    int rc = 0;
    size_t finalbytes;

	log_debug("DUKE: Copy %s -> %s", path, dest);

    startbytes = info->installed_bytes;
    finalbytes = startbytes + DECOMPSIZE;
    updatefunc = update;
    staticinfo = info;
    staticfinal = final;
    staticoption = current_option;

    snprintf(final, sizeof(final), "%s/duke3d.grp", dest);

	if ( update ) update(info, final, 0, 100, current_option);

    input = file_open(info, path, "rb");
    if ( input == NULL ) {
        return(-1);
    }

    if (!valid_archive(info, path, input))
    {
        log_fatal(_("This doesn't look like a known version of the MacSoft disc. Email ryan: icculus@clutteredmind.org"));
        return -1;
    }

	if ( update ) update(info, final, 0, 100, current_option);

    /* This is really scary... */
    inbuf = malloc(COMPSIZE);  /* compressed size */
    outbuf = malloc(DECOMPSIZE);  /* uncompressed size */
    if ((inbuf == NULL) || (outbuf == NULL))
    {
        free(inbuf);
        free(outbuf);
        log_fatal(_("Out of memory"));
        return -1;
    }

	if ( update ) update(info, final, 0, 100, current_option);

    while (br < COMPSIZE)
    {
        int total = 64 * 1024;
        if (total > COMPSIZE - br)
            total = COMPSIZE - br;
        rc = file_read(info, inbuf + br, total, input);
        if (rc != total)
        {
            free(inbuf);
            free(outbuf);
            log_fatal(_("i/o failure!"));
            return -1;
        }

        br += total;
        info->installed_bytes = startbytes + (br / 3);

    	if ( update ) {
            float pct = (((float) br) / ((float) COMPSIZE)) * 100.0f;
	        update(info, final, (int) (pct * 0.33f), 100, current_option);
    	}
    }

    file_close(info, input);

    startbytes = info->installed_bytes;

    decompress(inbuf, COMPSIZE, outbuf, DECOMPSIZE, 0x04);
    free(inbuf);

    startbytes = info->installed_bytes;

	output = file_open_install(info, final, "wb");
    if ( output == NULL ) {
        return(-1);
    }

    br = 0;
    while (br < DECOMPSIZE)
    {
        int total = 64 * 1024;
        if (total > DECOMPSIZE - br)
            total = DECOMPSIZE - br;
        rc = file_write(info, outbuf + br, total, output);
        if (rc != total)
        {
            free(outbuf);
            log_fatal(_("i/o failure!"));
            return -1;
        }

        br += total;
        info->installed_bytes = startbytes + (br / 3);

    	if ( update ) {
            float pct = (((float) br) / ((float) DECOMPSIZE)) * 100.0f;
	        update(info, final, (int) ((pct * 0.33f) + 66.0f), 100, current_option);
    	}
    }

    info->installed_bytes = finalbytes;

	if ( update ) {
	    update(info, final, 100, 100, current_option);
	}

	file_close(info, output);

    return DECOMPSIZE;
}



#ifdef DYNAMIC_PLUGINS
static
#endif
SetupPlugin duke_plugin = {
	"DUKE Archives Plugin",
	"1.0",
	"Ryan C. Gordon <icculus@clutteredmind.org>",
	1, {".DUKE"},
	DukeInitPlugin, DukeFreePlugin,
	DukeSize, DukeCopy
};

#ifdef DYNAMIC_PLUGINS
SetupPlugin *GetSetupPlugin(void)
{
	return &duke_plugin;
}
#endif
